from md.auction import Bid, Bidder, Divisibility
from md.auction_csv import to_number, encode_csv_solution
import re
from io import StringIO
from tabulate import tabulate

def parse_bidders(s):
    s = s.strip()
    lines = re.split('\n\\s*', s)
    bidders = []
    names = set()
    trader_id = 1
    last_n_goods = None
    for line in lines:
        if not line:
            continue

        # Optional bidder name is separated from bids by ':'
        if ':' in line:
            parts = line.split(':')
            if len(parts) > 2:
                raise ValueError('More than one colon in "{}"'.format(line))
            name = parts[0].strip()
            line = parts[1].strip()
        else:
            name = 'trader {}'.format(trader_id)
            trader_id += 1
        if name in names:
            raise ValueError('Two bidders with name {}'.format(name))
        else:
            names.add(name)
        bidder = Bidder(name)
        bidders.append(bidder)

        # XOR groups are separated by '|', bids by ',' and entries by ' '
        xor_groups = [x.strip() for x in line.split('|')]
        group_id = 1
        for group in xor_groups:
            bids = [x.strip() for x in group.split(',')]
            for bid_str in bids:
                items = bid_str.split()
                if items[0].lower() == 'd':
                    divisible = Divisibility.DIVISIBLE
                    items.pop(0)
                elif items[0].lower() == 'm':
                    divisible = Divisibility.MIXED
                    items.pop(0)
                else:
                    divisible = Divisibility.INDIVISIBLE

                n_goods = len(items)

                value = to_number(items[0])
                q = {}
                for i in range(1, len(items)):
                    good = 'good {}'.format(i)
                    q[good] = to_number(items[i])
                if len(bids) > 1:
                    xor_group = "xor_{}".format(group_id)
                else:
                    xor_group = None
                bidder.bids.append(Bid(value, q, divisible=divisible, xor_group=xor_group))

                if last_n_goods is not None:
                    if last_n_goods != n_goods:
                        raise ValueError(
                            "Not all bids have the same number of entries ({} and {})".format(last_n_goods, n_goods))
                last_n_goods = n_goods
            group_id += 1
    if len(bidders) < 2:
        raise ValueError("There must be at least 2 bidders.")
    return bidders

def to_txt(solution):
    si = StringIO()
    encode_csv_solution(solution, si, delimiter=',')

    data = []
    for row in si.getvalue().strip().split('\n'):
        data.append(row.strip().split(','))
    note = 'Pricing = {}, total surplus = {}, sum of payments = {}'.format(solution.rule, solution.surplus,
                                                                           solution.sum_payments())
    table = tabulate(data, headers='firstrow', tablefmt='rst')
    return table + '\n' + note

def encode_problem(problem):
    si = StringIO()
    goods = problem.list_goods()
    def write_bid(bid):
        if bid.divisibility == Divisibility.DIVISIBLE:
            si.write(" d")
        elif bid.divisibility == Divisibility.MIXED:
            si.write(" m")

        si.write(str(bid.v))
        for good in goods:
            if good in bid.q:
                si.write(" " + str(bid.q[good]))
            else:
                si.write(" 0")

    for bidder in problem.bidders:
        si.write(bidder.name)
        si.write(": ")
        xor_groups = bidder.xor_groups()

        first_group = True
        # Write xor groups
        for key, bid_ids in xor_groups.items():
            if not first_group:
                si.write("| ")
            first_bid = True
            for bid_id in bid_ids:
                if not first_bid:
                    si.write(", ")
                bid = bidder.bids[bid_id]
                write_bid(bid)
                first_bid = False
                first_group = False
        # Write bids not in xor groups.
        for bid in bidder.bids:
            if bid.xor_group is None:
                if not first_group:
                    si.write("| ")
                write_bid(bid)
                first_group = False

        si.write("\n")
    return si.getvalue()


from md.auction import *
import math
from random import Random
import itertools
from collections import defaultdict
import time

# Implementation of the payment rule described in the paper "Shapley value based pricing for auctions and exchanges"
# https://doi.org/10.1016/j.geb.2017.10.020

class Lindsay2018(PaymentRule):

    def __init__(self, max_perms_to_generate: int = 100000, max_perms_to_consider: int = 1024, seed: int = 7777):
        self.max_perms_to_generate = max_perms_to_generate
        self.max_perms_to_consider = max_perms_to_consider
        self.seed = seed

    def calc_surplus_shares(self, sol: Solution):

        #Used to give progress updates for slow problems.
        start_time = time.time()
        last_time = start_time

        winners = sol.winner_indexes()
        losers = sol.loser_indexes()

        permutations = []
        rnd = Random(self.seed)
        if math.factorial(len(winners)) > self.max_perms_to_generate:
            for i in range(self.max_perms_to_consider):
                perm = copy(winners)
                rnd.shuffle(perm)
                permutations.append(perm)
        else:
            permutations = list(itertools.permutations(winners))

        if len(permutations) > self.max_perms_to_consider:
            rnd.shuffle(permutations)
            permutations = permutations[0:self.max_perms_to_consider]

        # Stores running sum of shares for each winner.
        shares_sums = defaultdict(lambda: 0)

        # Cache surplus values to avoid repeatedly solving the same MIP problem.
        cache = {frozenset(winners): sol.surplus}

        problem = copy(sol.problem)
        auction = Auction()
        n_perms = len(permutations)

        for j, perm in enumerate(permutations):
            current_time = time.time()
            if current_time > last_time + 10:
                last_time = current_time
                print('Permutation {} of {}'.format(j, n_perms))
            last_surplus = 0
            problem.bidders = [sol.problem.bidders[i] for i in losers]
            for position, i in enumerate(perm):
                problem.bidders.append(sol.problem.bidders[i])

                key = frozenset(perm[0:position + 1])
                if key in cache.keys():
                    surplus = cache[key]
                else:
                    surplus = auction.surplus_determination(problem)
                    cache[key] = surplus

                share = surplus - last_surplus
                shares_sums[i] = shares_sums[i] + share
                last_surplus = surplus


        for i in winners:
            surplus_share = shares_sums[i] / n_perms
            sol.populate_surplus_and_payment(sol.problem.bidders[i], surplus_share)

        sol.rule = 'lindsay2018'
        current_time = time.time()
        if current_time > start_time + 10:
            n_winners = len(winners)
            n_bidders = n_winners + len(losers)
            solve_time = round(current_time-start_time)
            print('{} bidders and {} winners.  Calculating prices took {} seconds with {} permutations.'.format(n_bidders, n_winners,solve_time, n_perms))


def get_rule(rule_name,  max_perms_to_consider: int = 1024):
    rule_name = rule_name.lower().strip()  # ignore case
    if rule_name == 'lindsay2018':
        rule = Lindsay2018(max_perms_to_consider=max_perms_to_consider)
    elif rule_name == 'vcg':
        rule = VCG()
    elif rule_name == 'pab':
        rule = PaB()
    else:
        raise ValueError('Unrecognised payment rule "{}".'.format(rule_name))
    return rule



from md.auction import Bid, Bidder, Problem, Divisibility
from json import JSONEncoder


class ObjectEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, Divisibility):
            return o.value
        dct = o.__dict__
        # Remove None keys with value None
        return {k: v for k, v in dct.items() if v is not None}


def decode_bid(dct):
    if 'divisible' in dct and 'divisibility' in dct:
        raise ValueError("Has both divisible and divisibility")

    for k in ['label', 'winning', 'xor_group']:
        if k not in dct:
            dct[k] = None

    divisible = Divisibility.INDIVISIBLE
    #Old api
    if 'divisible' in dct:
        if dct['divisible']:
            divisible = Divisibility.DIVISIBLE
        else:
            divisible = Divisibility.INDIVISIBLE

    #Newer api
    if 'divisibility' in dct:
        divisible = Divisibility(dct['divisibility'])

    return Bid(dct['v'], dct['q'], label=dct['label'], winning=dct['winning'], xor_group=dct['xor_group'],
               divisible=divisible)


def decode_bidder(dct):
    b = Bidder(dct['name'])
    for bid in dct['bids']:
        b.bids.append(decode_bid(bid))

    return b


def decode_problem(dct):
    bidders = []
    for bidder in dct['bidders']:
        bidders.append(decode_bidder(bidder))
    for k in ['description', 'goods']:
        if k not in dct:
            dct[k] = None
    if 'free_disposal' not in dct:
        dct['free_disposal'] = True
    problem = Problem(bidders=bidders, description=dct['description'], goods=dct['goods'],
                   free_disposal=dct['free_disposal'])
    problem.validate()
    return problem



from mip import Model, xsum, maximize, BINARY, CONTINUOUS
from collections import defaultdict
from copy import copy
from enum import Enum, unique


@unique
class Divisibility(Enum):
    INDIVISIBLE = 0
    DIVISIBLE = 1
    MIXED = 2


class Bid:
    def __init__(self, v, q, xor_group=None, winning=None, label=None,
                 divisible: Divisibility = Divisibility.INDIVISIBLE):
        if not isinstance(divisible, Divisibility):
            raise TypeError
        self.v = v
        self.q = q
        self.xor_group = xor_group
        self.winning = winning
        self.label = label
        self.divisibility = divisible


class Bidder:

    def __init__(self, name):
        self.name = name
        self.bids = []

    def add_bid(self, v, q, xor_group=None, label=None, divisible=Divisibility.INDIVISIBLE):
        b = Bid(v, q, label=label, divisible=divisible)
        if xor_group is not None:
            b.xor_group = xor_group
        self.bids.append(b)
        return self

    def bid_ids(self):
        return range(len(self.bids))

    def xor_groups(self):
        groups = defaultdict(list)
        for i, bid in enumerate(self.bids):
            if bid.xor_group is not None:
                groups[bid.xor_group].append(i)
        return groups


class Problem:

    def __init__(self, description=None, goods=None, bidders=None, free_disposal=True):
        if bidders is None:
            bidders = []
        self.description = description
        self.goods = goods
        self.free_disposal = free_disposal
        self.bidders = bidders

    def __copy__(self):
        return Problem(description=self.description, goods=self.goods, bidders=self.bidders,
                       free_disposal=self.free_disposal)

    def add_bidder(self, bidder):
        self.bidders.append(bidder)
        return self

    # Returns an order list of the goods.
    def list_goods(self):
        s = set()
        for bidder in self.bidders:
            for bid in bidder.bids:
                for good in bid.q.keys():
                    s.add(good)
        return sorted(s)

    # Returns a list of bidders' names
    def bidder_names(self):
        names = []
        for bidder in self.bidders:
            names.append(bidder.name)

        return names

    # Raises an exception if an inconsistency is detected.
    def validate(self):
        # Check bids in the same xor group have the same divisibility.
        # Also check for repeated bidder names.
        bidder_names = set()
        for bidder in self.bidders:
            if bidder.name in bidder_names:
                raise ValueError("Repeated bidder name: {}".format(bidder.name))

            bidder_names.add(bidder.name)
            xorgroup2type = {}
            for bid in bidder.bids:
                if bid.xor_group:
                    if bid.xor_group in xorgroup2type:
                        if bid.divisibility != xorgroup2type[bid.xor_group]:
                            raise ValueError(
                                "Bids in same xor group should have the same divisibility: {}, {}".format(bidder.name, bid.xor_group))
                    else:
                        xorgroup2type[bid.xor_group] = bid.divisibility

class Solution:

    def __init__(self, rule=None, surplus=None, surplus_shares=None, payments=None, problem=None):
        if payments is None:
            payments = {}
        if surplus_shares is None:
            surplus_shares = {}
        self.rule = rule
        self.surplus = surplus
        self.surplus_shares = surplus_shares
        self.payments = payments
        self.problem = problem

    def winner_indexes(self):
        winners = []
        for i, bidder in enumerate(self.problem.bidders):
            for bid in bidder.bids:
                if bid.winning > 0:
                    winners.append(i)
                    break
        return winners

    def loser_indexes(self):
        losers = []
        winners = self.winner_indexes()
        for i in range(len(self.problem.bidders)):
            if i not in winners:
                losers.append(i)
        return losers

    def populate_surplus_and_payment(self, bidder, surplus_share):
        self.surplus_shares[bidder.name] = surplus_share

        # payment is sum of winning bids minus surplus share.
        payment = -surplus_share
        for bid in bidder.bids:
            payment += bid.v * bid.winning
        self.payments[bidder.name] = payment

    def sum_payments(self):
        return sum(self.payments.values())


class Auction:

    def build_model(self, p):
        m = Model('winner_determination')
        m.verbose = 0
        I = range(len(p.bidders))

        mixed_xor_groups = set()

        def add_bid_var(i, j):
            name = 'bid[{}][{}]'.format(i, j)
            bid = p.bidders[i].bids[j]
            if bid.divisibility == Divisibility.INDIVISIBLE:
                return m.add_var(var_type=BINARY, name=name)
            else:
                if bid.divisibility == Divisibility.MIXED and bid.xor_group is not None:
                    mixed_xor_groups.add(bid.xor_group)

                return m.add_var(var_type=CONTINUOUS, lb=0, ub=1, name=name)

        # Create variables
        winning = [[add_bid_var(i, j) for j in p.bidders[i].bid_ids()] for i in I]

        # Setup objective
        m.objective = maximize(
            xsum(p.bidders[i].bids[j].v * winning[i][j] for i in I for j in p.bidders[i].bid_ids()))

        # Add supply-demand constraints
        for good in p.list_goods():
            lhs = xsum(winning[i][j] * p.bidders[i].bids[j].q[good] for i in I for j in
                       p.bidders[i].bid_ids() if good in p.bidders[i].bids[j].q)
            if p.free_disposal:
                m += lhs >= 0, 'supply is at least demand'
            else:
                m += lhs == 0, 'supply equals demand'

        # Add XOR constraints
        for i in I:
            for group_name, bid_ids in p.bidders[i].xor_groups().items():
                if group_name in mixed_xor_groups:
                    # Need to setup a variable.
                    name = 'xorgroup_{}'.format(group_name)
                    var = m.add_var(var_type=BINARY, name=name)
                    m += xsum(winning[i][j] for j in bid_ids) == var, 'xor[{}][{}]'.format(i, group_name)
                elif len(bid_ids) > 1:
                    m += xsum(winning[i][j] for j in bid_ids) <= 1, 'xor[{}][{}]'.format(i, group_name)
        return m, I, winning

    def winner_determination(self, p):
        rt = Solution(problem=p)

        m, I, winning = self.build_model(p)
        m.optimize()
        for i in I:
            for j in p.bidders[i].bid_ids():
                if winning[i][j].x > 0:
                    p.bidders[i].bids[j].winning = winning[i][j].x
                else:
                    p.bidders[i].bids[j].winning = 0

        rt.surplus = m.objective_value
        p.goods = p.list_goods()
        return rt

    # Determines and returns surplus without populating the winning bids.  Intended
    # to be used when calculating vcg payment etc.
    def surplus_determination(self, p):
        m, I, winning = self.build_model(p)
        m.optimize()
        return m.objective_value


class PaymentRule:
    def calc_surplus_shares(self, solution: Solution):
        pass


class PaB(PaymentRule):
    def calc_surplus_shares(self, sol: Solution):
        for i in sol.winner_indexes():
            bidder = sol.problem.bidders[i]
            sol.populate_surplus_and_payment(bidder, 0)

        sol.rule = 'pab'


class VCG(PaymentRule):
    def calc_surplus_shares(self, sol: Solution):
        # Work with a shallow copy of the problem, the list of bidders will be replaced in the loop below.
        problem = copy(sol.problem)
        auction = Auction()
        for i in sol.winner_indexes():
            problem.bidders = copy(sol.problem.bidders)
            bidder = problem.bidders.pop(i)
            surplus_without = auction.surplus_determination(problem)
            surplus_share = sol.surplus - surplus_without
            sol.populate_surplus_and_payment(bidder, surplus_share)

        sol.rule = 'vcg'



from md.auction import Bid, Bidder, Divisibility, Problem
import csv
import chardet as chardet
import os

HEADERS = ('name', 'xor_group', 'label', 'divisible', 'value')


def to_number(s):
    try:
        x = int(s)
    except ValueError:
        x = float(s)
    return x

def get_file_extension(filename):
    _, file_extension = os.path.splitext(filename)
    return file_extension.lower()

def decode_csv_bid(dct, goods):
    for k in ['label', 'divisible', 'xor_group']:
        if k not in dct:
            dct[k] = None
        elif dct[k] is not None:
            if dct[k].strip() == '':
                dct[k] = None

    # Set xor_group to None in case of empty string.
    if dct['xor_group'] is not None:
        if dct['xor_group'].strip() == '':
            dct['xor_group'] = None
    q = {}
    for good in goods:
        if good in dct.keys():
            s = dct[good].strip()
            # Ignore empty strings
            if s:
                q[good] = to_number(s)
    v = to_number(dct['value'])
    if dct['divisible'] == '1':
        divisible = Divisibility.DIVISIBLE
    elif dct['divisible'] == '2':
        divisible = Divisibility.MIXED
    else:
        divisible = Divisibility.INDIVISIBLE
    return Bid(v, q, label=dct['label'], xor_group=dct['xor_group'], divisible=divisible)

def decode_csv_problem(reader: csv.DictReader):
    bidders = decode_csv_bidders(reader)
    problem = Problem(bidders=bidders)
    problem.validate()
    return problem

def decode_csv_bidders(reader: csv.DictReader):
    # Check for duplicate column names
    n_columns = len(reader.fieldnames)
    if n_columns != len(set(reader.fieldnames)):
        raise ValueError(str(reader.fieldnames))

    goods = [good for good in reader.fieldnames if good not in HEADERS]
    bidders = []
    name2bidder = {}
    for row in reader:
        if n_columns != len(row):
            raise ValueError(row)
        name = row['name']
        if name not in name2bidder.keys():
            bidder = Bidder(name)
            name2bidder[name] = bidder
            bidders.append(bidder)
        else:
            bidder = name2bidder[name]
        bidder.bids.append(decode_csv_bid(row, goods))

    return bidders


def encode_csv_solution(solution, csv_file, delimiter=','):
    goods = solution.problem.list_goods()
    extra = ['winning', 'surplus share', 'payment']
    headers = list(HEADERS) + goods + extra
    writer = csv.DictWriter(csv_file, fieldnames=headers, delimiter=delimiter)
    writer.writeheader()

    for bidder in solution.problem.bidders:
        # First write the bids.
        for bid in bidder.bids:
            row = {'name': bidder.name, 'value': bid.v}
            if bid.xor_group:
                row['xor_group'] = bid.xor_group
            if bid.label:
                row['label'] = bid.label
            if bid.divisibility:
                v = bid.divisibility.value
                if v == 0:
                    v = None
                row['divisible'] = v
            for good in bid.q.keys():
                row[good] = bid.q[good]
            row['winning'] = bid.winning
            writer.writerow(row)

        # Second write the surplus shares row.
        row = {'name': bidder.name,
               'surplus share': solution.surplus_shares.get(bidder.name),
               'payment': solution.payments.get(bidder.name)}
        writer.writerow(row)

def encode_csv_problem(problem, csv_file, delimiter=','):
    goods = problem.list_goods()
    headers = list(HEADERS) + goods
    writer = csv.DictWriter(csv_file, fieldnames=headers, delimiter=delimiter)
    writer.writeheader()

    for bidder in problem.bidders:
        # First write the bids.
        for bid in bidder.bids:
            row = {'name': bidder.name, 'value': bid.v}
            if bid.xor_group:
                row['xor_group'] = bid.xor_group
            if bid.label:
                row['label'] = bid.label
            if bid.divisibility:
                v = bid.divisibility.value
                if v == 0:
                    v = None
                row['divisible'] = v
            for good in bid.q.keys():
                row[good] = bid.q[good]
            writer.writerow(row)

def file2reader(f):
    # Use chardet to infer the encoding of the csv file.
    result = chardet.detect(f.read())
    encoding = result['encoding']
    f.seek(0)

    fstring = f.read().decode(encoding)

    lines = fstring.splitlines()
    header = [h.strip() for h in lines[0].split(',')]
    lines.pop(0)
    return csv.DictReader(lines, fieldnames=header)

def string2reader(s):
    lines = s.splitlines()
    header = [h.strip() for h in lines[0].split(',')]
    lines.pop(0)
    return csv.DictReader(lines, fieldnames=header)